import torch
from torch.utils import tensorboard as tb
import cv2
import numpy as np
from ..transforms import imagenet
from ..mappings.color import get_palette


class SummaryWriter(tb.SummaryWriter):
    """Extension of Summary Writer for convenient common uses."""

    def __init__(self, *args, rgb_transform=None, **kwargs):
        """
        Parameters
        ----------
        rgb_transform : str, torchvision.transforms.*
            Used when rgb images need to be logged. Transform applied to
            rgb images when logged.
        """
        super().__init__(*args, **kwargs)

        self.rgb_transform = self._parse_rgb_transform(rgb_transform)

    def _parse_rgb_transform(self, rgb_transform):
        if rgb_transform is None:
            try:
                return self.rgb_transform
            except AttributeError:
                return lambda x: x  # callable identity function

        if callable(rgb_transform):
            return rgb_transform

        if isinstance(rgb_transform, str):
            rgb_transform = rgb_transform.lower()
            if rgb_transform == "imagenet":
                return imagenet.Unnormalize()

        raise NotImplementedError(f"Cannot parse rgb_transform {rgb_transform}")

    def add_rgb(
        self,
        tag,
        rgbs,
        global_step=None,
        walltime=None,
        dataformats="CHW",
        *,
        rgb_transform=None,
        n_images=1,
    ):
        """Applies a transform and adds image to log

        A common scenario is when you only have a normalized image (say, by
        ImageNet's mean and stddev) and you want to log it to tensorboard.

        In this case, it may be more convenient to set:
            rgb_transform="imagenet"
        in the constructor.

        Paramters
        ---------
        rgbs : torch.Tensor
            (B, 3, H, W) Image data.  See ``rgb_transform`` argument.
        rgb_transform : str or callable
            Transform to apply to the rgb data. If not provided, defaults to
            the transform provided in __init__
        n_images : int
            Maximum number of images to add.
        """

        rgb_transform = self._parse_rgb_transform(rgb_transform)

        rgbs = rgb_transform(rgbs)
        rgbs = rgbs.cpu().numpy()
        rgbs = np.clip(rgbs, 0, 1)
        rgbs = np.transpose(rgbs, (0, 2, 3, 1))  # (B, H, W, 3)
        rgbs = (rgbs * 255).astype(np.uint8)

        for i, rgb in zip(range(n_images), rgbs):
            self.add_image(
                f"{tag}/{i}",
                rgb,
                global_step=global_step,
                walltime=walltime,
                dataformats="HWC",
            )

    def add_ss(
        self,
        tag,
        rgbs,
        preds,
        targets,
        global_step=None,
        walltime=None,
        dataformats="CHW",
        *,
        rgb_transform=None,
        n_images=1,
        palette="ade20k",
        offset=0,
    ):
        """Add a semantic segmentation image and it's pairing input montage.

        ``self.add_rgb``'s documentation applies to the ``rgbs`` input here.

        TODO: more control over which image to show

        Parameters
        ----------
        tag : str
            Data identifier
        rgbs : torch.Tensor
            (B, 3, H, W) Image data.  See ``rgb_transform`` argument.
        preds : torch.Tensor
            (B, C, H, W) Predicted semantic segmentation data. This method will
            argmax over the C dimension.
        targets : torch.Tensor
            (B, H, W) Indexed ground truth data.
        rgb_transform : str or callable
            Transform to apply to the rgb data. If not provided, defaults to
            the transform provided in __init__. Expects data to be in range
            [0, 1] after transform.
        n_images : int
            Maximum number of images to add.
        offset : int
            Add this to the pred and target index into the colormap.
            A common value might be 1 if your network isn't using a
            background class.
        """

        # Input validation
        if dataformats != "CHW":
            raise NotImplementedError("TODO: allow other dataformats")
        if isinstance(palette, str):
            palette = get_palette(palette)
        elif isinstance(palette, np.ndarray):
            # Do nothing, TODO: maybe add some assertions here
            pass
        else:
            raise NotImplementedError(
                f"Don't know how to handle palette type {type(palette)}"
            )
        n_colors = len(palette)
        rgb_transform = self._parse_rgb_transform(rgb_transform)

        for i, rgb in enumerate(rgbs):
            rgbs[i] = rgb_transform(rgb)

        # Get the most likely class from the network's logits/softmax/whatever.
        preds = torch.argmax(preds, dim=1)  # (B, H, W)

        # Move all the data to cpu
        rgbs = rgbs.cpu().numpy()
        preds = preds.cpu().numpy().astype(np.int)
        targets = targets.cpu().numpy().astype(np.int)

        # General cleanup operations
        rgbs = np.clip(rgbs, 0, 1)
        rgbs = np.transpose(rgbs, (0, 2, 3, 1))  # (B, H, W, 3)
        rgbs = (rgbs * 255).astype(np.uint8)

        preds += offset
        targets += offset

        _, h, w, _ = rgbs.shape

        # Iterate over exemplars and log
        # Note: ``zip`` limits iterations to the shortest input iterator.
        for i, rgb, pred, target in zip(range(n_images), rgbs, preds, targets):
            # Resize pred and target
            pred = cv2.resize(pred, (w, h), interpolation=cv2.INTER_NEAREST)
            target = cv2.resize(target, (w, h), interpolation=cv2.INTER_NEAREST)

            pred[pred >= n_colors] = 0
            pred[pred < 0] = 0

            target[target >= n_colors] = 0
            target[target < 0] = 0

            # Apply a color palette
            color_pred = palette[pred]
            color_target = palette[target]
            color_target[target < 0] = 0

            # Horizontally combine all three into a single image
            montage = np.concatenate((color_target, rgb, color_pred), axis=1).astype(
                np.uint8
            )

            # Log the montage to tensorboard
            self.add_image(
                f"{tag}/{i}",
                montage,
                global_step=global_step,
                walltime=walltime,
                dataformats="HWC",
            )



# -*- coding: utf-8 -*-

"""Top-level package for Pugh Torch."""

__author__ = "Brian Pugh"
__email__ = "bnp117@gmail.com"
# Do not edit this string manually, always use bumpversion
# Details in CONTRIBUTING.md
__version__ = "0.2.0"


def get_module_version():
    return __version__


from .example import Example  # noqa: F401



import torch
from torchvision import transforms
from PIL import Image
import numpy as np

__all__ = [
    "Normalize",
    "Unnormalize",
]


class Normalize:
    """Applies standard ImageNet normalization to an RGB image"""

    def __init__(
        self,
    ):
        self.to_tensor = transforms.ToTensor()
        self.normalize = transforms.Normalize(
            mean=[0.485, 0.456, 0.406],
            std=[0.229, 0.224, 0.225],
        )

    def __call__(self, img):
        """
        Parameters
        ----------
        img : numpy.ndarray or PIL.Image or torch.Tensor
            If numpy.ndarray or PIL.Image, expects (H,W,3).
            If torch.Tensor, expects (3, H, W)
            Expects values to be in range [0, 1], NOT [0,255]
            Channels are in RGB order.

        Returns
        -------
        torch.Tensor
            (C, H, W)
        """

        if not isinstance(img, torch.Tensor):
            # Need to convert to torch.Tensor first
            if isinstance(img, np.ndarray):
                if img.max() > 1 or np.issubdtype(img.dtype, np.integer):
                    raise ValueError(
                        "Image must be a float in range [0,1] and in RGB order prior to transform"
                    )
            img = self.to_tensor(img)  # Converts to tensor, and permutes to (C, H, W)
        return self.normalize(img)  # Expects (C, H, W)


class Unnormalize:
    """Reverses standard ImageNet normalization"""

    def __init__(
        self,
    ):
        self.unnormalize = transforms.Normalize(
            mean=[-0.485 / 0.229, -0.456 / 0.224, -0.406 / 0.225],
            std=[1 / 0.229, 1 / 0.224, 1 / 0.225],
        )

    def __call__(self, img):
        """
        Parameters
        ----------
        img : torch.Tensor
            (3, H, W) normalized tensor (should be apprixmately in range [-1, 1])

        Returns
        -------
        img : torch.Tensor
            (3, H, W) unnormalized tensor (should be approximately in range [0, 1])
        """

        return self.unnormalize(img)



import torch
import torch.nn.functional as F


def hetero_cross_entropy(
    preds,
    targets,
    availables,
    *,
    super_index=0,
    ignore_index=-100,
):
    """Cross Entropy Loss for Heterogenous Datasets.

    Basically a cross-entropy-loss for when you are training on multiple
    datasets where some of the datasets may have some of the classes labeled
    as a single superset (for example, background).

    Wordy logic:
        1. Lets say Dataset A has classes ["cat", "dog", "bird"].
        2. Lets say Dataset B has classes ["cat",              , "other"].
           i.e. Dataset B is missing class "dog" and "bird", and those pixels
           fall into the "other" category.
        3. For Dataset A, ``hetero_cross_entropy`` is exactly the same as
           normal cross entropy loss.
        4. For Dataset B, a "dog" will be labeled into the "other" category,
           but we don't want to penalize the network for predicting dog,
           because it could be correct.  However, we absolutely know that
           the network should not be predicting "cat".
        5. To solve this on dataset B, we apply cross entropy loss over the
           pixels not in the "other" class in the target. In addition to this,
           we want to maximize the probability that the network guessed either
           "dog" or "bird" for the pixels labeled as "other" in the target.

    Described in "Combining Heterogeneously Labeled Datasets For Training
    Segmentation Networks"
        https://arxiv.org/pdf/1807.08935.pdf

    TODO: this could be extended to multiple supersets, but the API might
    get complicated, and I personally don't have a use yet.

    Parameters
    ----------
    preds : torch.Tensor
        (N, C, ...) where C is the number of classes.
        Typically this is the **logits** coming out of your network.
    targets : torch.Tensor
        (N, ...) T
    availables : torch.Tensor
        (N, C) One-hot vector of which classes are availble from the dataset
        for this exemplar.
    ignore_index : int
        Pixels with this target value will be not contribute to the loss what
        so ever. Typically this value might be what you pad the target
        segmentation image with if you are performing random crops.
        May be negative.
    super_index : int
        Pixels with this target value will be encouraged to be of classes not
        in the dataset it came from. Typically this is a "background", "unknown",
        or "other" class.
        May be negative.

    Returns
    -------
    torch.Tensor
        Scalar loss value
    """

    # Make sure all the batch sizes are the same
    assert len(preds) == len(targets) == len(availables)

    super_hots = (
        ~availables.bool()
    )  # (B, C) where 1 means that class is a part of the superclass

    # Initialize the two components of our loss on the correct device
    ce_loss = preds.new([0]).float()
    ce_loss.requires_grad_()  # Just in case we don't actually add to this loss, ``backward()`` will still work
    super_loss = preds.new([0]).float()

    # Iterate over the batch dimension because the shapes are going to get jagged.
    for pred, target, super_hot in zip(preds, targets, super_hots):
        # pred - (C, H, W) float
        # target - (H, W) long
        # super_hot - (C,) bool
        inbound_mask = target != ignore_index  # (H, W)
        super_mask = target == super_index
        ce_mask = inbound_mask * ~super_mask

        # Apply CE to inbound mask, excluding the super_index
        if torch.any(ce_mask):
            ce_pred = pred[:, ce_mask]  # (C, n_valid)
            ce_target = target[ce_mask]  # (n_valid)
            ce_pred = ce_pred.transpose(0, 1)  # (n_valid, C)
            ce_loss = ce_loss + F.cross_entropy(ce_pred, ce_target)

        if torch.any(super_mask):
            super_pred = pred[:, super_mask]  # (C, n_pix)

            # Compute the softmax over all classes
            super_pred_exp = torch.exp(super_pred)
            super_loss = super_loss - torch.mean(
                torch.log(super_pred_exp[super_hot].sum(dim=0))
                - torch.log(super_pred_exp.sum(dim=0))
            )

    return ce_loss + super_loss
